import os
from os.path import dirname
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import sentinel

from pytest import fixture
from pytest import raises

from qq import Application
from qq import Context
from qq.plugins.settings import SettingsBasedPlugin
from qq.plugins.settings import SettingsInicjator
from qq.plugins.settings import SettingsPlugin
from qq.plugins.settings import _import
from qq.testing import PluginFixtures

PREFIX = "qq.plugins.settings"


class TestSettingsPlugin(PluginFixtures):
    MODULE = sentinel.settings_module

    @fixture
    def plugin(self):
        plugin = SettingsPlugin(self.MODULE)
        plugin.init("settings")
        return plugin

    @fixture
    def mstring_dict(self):
        return {}

    @fixture
    def mapp(self):
        return MagicMock()

    @fixture
    def mimport(self):
        with patch(f"{PREFIX}._import") as mock:
            yield mock

    def test_application(self, plugin, mapplication, mapp):
        """
        .enter should add settings and paths to the application.
        """
        context = MagicMock()
        context.globals = {"settings": sentinel.settings}
        assert plugin.enter(context) == sentinel.settings

    def test_import(self, plugin):
        assert _import("os") == os

    def test_start(self, plugin, mapp, mimport):
        """
        .start should gather settings for startpoint and push the data into
        configurator.
        """
        mapp.startpoint = "myapp"

        result = plugin.start(mapp)

        mimport.assert_called_once_with(self.MODULE)
        mimport.return_value.myapp.assert_called_once_with()

        assert result == mimport.return_value.myapp.return_value


class TestSettingsBasedPlugin:
    PLUGINKEY = "ksdyuwyta"

    @fixture
    def app(self):
        app = Application()
        app.globals["settings"] = {self.PLUGINKEY: sentinel.settings}
        return app

    @fixture
    def context(self, app):
        context = Context(app)
        context.values["settings"] = {self.PLUGINKEY: sentinel.settings}
        return context

    @fixture
    def plugin(self):
        plugin = SettingsBasedPlugin()
        plugin.init(self.PLUGINKEY)
        return plugin

    def test_get_my_settings_when_application(self, app, plugin):
        assert plugin.get_my_settings(app) == sentinel.settings

    def test_get_my_settings_when_context(self, context, plugin):
        assert plugin.get_my_settings(context) == sentinel.settings


class TestSettingsInicjator:
    PLUGINKEY = "ksdyuwyta"

    @fixture
    def app(self):
        app = Application()
        app.globals["settings"] = {self.PLUGINKEY: sentinel.settings}
        return app

    def test_flow(self, app):
        context = {
            SettingsPlugin.key: {self.PLUGINKEY: sentinel.pluginsettings}
        }
        inicjator = SettingsInicjator(self.PLUGINKEY)
        inicjator.init(context, None)
        result = inicjator.start()
        assert result == sentinel.pluginsettings



class QQError(RuntimeError):
    pass


class ApplicationNotStartedError(QQError):
    pass


class AlreadyStartedError(QQError):
    pass


class PluginLacksOfKeyError(QQError):
    pass


class WrongKeyForPluginError(QQError):
    pass


class PluginKeyAlreadyDefinedError(QQError):
    pass



"""
This module provide tests for the main functionality of the Quack Quack.
"""

from pytest import raises

from qq import Application
from qq import Context
from qq.errors import ApplicationNotStartedError
from qq.injectors import ContextInicjator
from qq.injectors import Inicjator
from qq.injectors import SetApplication
from qq.injectors import SetInicjator
from qq.plugin import Plugin
from qq.plugins import SettingsPlugin
from qq.plugins.types import Settings


class SamplePlugin(Plugin):
    key = "first"

    def __init__(self, value: str):
        super().__init__()
        self.value = value

    def start(self, application: Application):
        return {
            "data": self.value,
        }

    def enter(self, context: Context):
        return context.globals[self.key]


class SecondSamplePlugin(Plugin):
    def __init__(self, value: str):
        super().__init__()
        self.value = value

    def start(self, application: Application):
        return {
            "data": self.value,
        }

    def enter(self, context: Context):
        return context.globals[self.key]


class MyApplication(Application):
    def create_plugins(self):
        self.plugins(SamplePlugin("one"))
        self.plugins["second"] = SecondSamplePlugin("two")


class AsyncContextInicjator(Inicjator):
    def __init__(self, key: str):
        self.key = key

    async def start(self):
        return self.context[self.key]

    async def finish(self, er=None):
        pass


application = MyApplication()
application.start()


def test_normal_function():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", ContextInicjator("second"))
    def mytest(first, second):
        return first, second

    assert mytest() == ({"data": "one"}, {"data": "two"})


async def test_async_function():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", ContextInicjator("second"))
    async def mytest(first, second):
        return first, second

    assert await mytest() == ({"data": "one"}, {"data": "two"})


def test_set_only_application():
    @SetApplication(application)
    def mytest():
        return "one"

    assert mytest() == "one"


def test_set_only_inicjator():
    @SetInicjator("first", ContextInicjator("first"))
    def mytest():
        return "one"

    with raises(ApplicationNotStartedError):
        mytest()


def test_error_function():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    def mytest(first):
        raise RuntimeError("X")

    with raises(RuntimeError):
        mytest()


def test_dependency_injection():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", ContextInicjator("second"))
    def mytest(first, second):
        return first, second

    assert mytest("one", "two") == ("one", "two")


def test_dependency_injection_kwargs():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", ContextInicjator("second"))
    def mytest(first, second):
        return first, second

    assert mytest(second="two", first="one") == ("one", "two")


def test_repr():
    def mytest():
        pass

    assert repr(SetApplication(application)(mytest)) == repr(mytest)


async def test_async_inicjator():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", AsyncContextInicjator("second"))
    async def mytest(first, second):
        return first, second

    assert await mytest() == ({"data": "one"}, {"data": "two"})


async def test_async_inicjator_error():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", AsyncContextInicjator("second"))
    async def mytest(first, second):
        raise RuntimeError("x")

    with raises(RuntimeError):
        await mytest()


async def test_dependency_injection_async():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", ContextInicjator("second"))
    async def mytest(first, second):
        return first, second

    assert await mytest("one", "two") == ("one", "two")


async def test_dependency_injection_async_kwargs():
    @SetApplication(application)
    @SetInicjator("first", ContextInicjator("first"))
    @SetInicjator("second", ContextInicjator("second"))
    async def mytest(first, second):
        return first, second

    assert await mytest(second="two", first="one") == ("one", "two")



from sqlalchemy.orm import Session

from qq.injectors import Inicjator
from qq.plugins.settings import TESTS_KEY
from qq.plugins.settings import SettingsPlugin
from qq.plugins.types import Settings


class SesssionInicjator(Inicjator):
    def __init__(self, key: str):
        self.key = key

    def start(self):
        self.session: Session = self.context[self.key]
        self.settings: Settings = self.context[SettingsPlugin.key]
        return self.session


class TransactionInicjator(SesssionInicjator):

    def finish(self, er=None):
        if er:
            return self.session.rollback()
        elif self.settings.get(TESTS_KEY, False):
            return self.session.flush()
        else:
            return self.session.commit()


class TransactionInicjatorAsync(SesssionInicjator):

    async def finish(self, er=None):
        await super().finish(er)



from celery import Celery

from qq.application import Application
from qq.plugins.settings import SettingsBasedPlugin


class CeleryPlugin(SettingsBasedPlugin):
    key = "celery"

    def __init__(self, celeryapp: Celery):
        super().__init__()
        self.celeryapp = celeryapp

    def start(self, application: Application):
        settings = self.get_my_settings(application)
        self.celeryapp.conf.update(settings)



from typing import Any

from qq.context import Context
from qq.types import Application
from qq.types import Plugin as PluginType
from qq.types import PluginKey


class Plugin(PluginType):
    key: PluginKey = None

    def init(self, key: PluginKey):
        """
        Initialize the plugin during creation.
        key - key which is used in the Application.plugins dict for this plugin.
        """
        self.key = key

    def start(self, application: Application) -> Any:
        """
        This method will be called at the start of the Application. It will be
        called only once and the result will be set in the Application.globals.
        """

    def enter(self, context: Context) -> Any:
        """
        This method will be called when the Application will be used as context
        manager, but only when the plugin will be called. This is the enter phase.
        Result will be set in the Context dict with the self.key as the key in
        that dict.
        """

    def exit(self, context: Context, exc_type, exc_value, traceback):
        """
        This method will be called when the Application will be used as context
        manager. This is the exit phase.
        """



from unittest.mock import MagicMock

from pytest import fixture
from pytest import raises

from qq.plugins.sqlalchemy.exceptions import SettingMissing
from qq.plugins.sqlalchemy.plugin import SqlAlchemyPlugin
from qq.plugins.sqlalchemy.plugin import SqlAlchemyPluginAsync

PREFIX = "qq.plugins.sqlalchemy.plugin"


class TestSqlAlchemyPlugin:
    @fixture
    def mapp(self):
        return MagicMock()

    @fixture
    def mcreate_engine(self, mocker):
        return mocker.patch(f"{PREFIX}.create_engine")

    @fixture
    def mmake_url(self, mocker):
        return mocker.patch(f"{PREFIX}.make_url")

    @fixture
    def msession(self, mocker):
        return mocker.patch(f"{PREFIX}.Session")

    @fixture
    def plugin(self):
        plugin = SqlAlchemyPlugin()
        plugin.init("dbname")
        return plugin

    @fixture
    def mget_my_settings(self, mocker, plugin):
        mock = mocker.patch.object(plugin, "get_my_settings")
        mock.return_value = {
            "url": "sqlite:///tmp.first.db",
            "options": {"optionkey": "option value"},
        }
        return mock

    @fixture
    def metadata(self):
        return MagicMock()

    def test_start(self, plugin, mapp, mcreate_engine, mget_my_settings):
        """
        .start should create proper sqlalchemy engine.
        """
        plugin.start(mapp)

        mcreate_engine.assert_called_once_with(
            "sqlite:///tmp.first.db", optionkey="option value"
        )

    def test_dbname(
        self,
        plugin,
        mapp,
        mcreate_engine,
        mget_my_settings,
    ):
        """
        .dbname should return name of the database made from the db url
        """
        plugin.start(mapp)

        assert plugin.dbname == "tmp.first.db"

    def test_validate_settings(self, plugin, mapp, mget_my_settings):
        """
        Starting plugin should raise an error when settings are not properly
        configured (missing url)
        """
        del mget_my_settings.return_value["url"]
        with raises(SettingMissing):
            plugin.start(mapp)

    def test_enter(self, plugin, msession):
        """
        .enter should create new database session and assign it to the context
        """
        mcontext = MagicMock()
        plugin.engine = MagicMock()

        assert plugin.enter(mcontext) == msession.return_value
        msession.assert_called_once_with(plugin.engine, expire_on_commit=False)


class TestSqlAlchemyPluginAsync:
    @fixture
    def mapp(self):
        return MagicMock()

    @fixture
    def masync_session(self, mocker):
        return mocker.patch(f"{PREFIX}.AsyncSession")

    @fixture
    def mcreate_async_engine(self, mocker):
        return mocker.patch(f"{PREFIX}.create_async_engine")

    @fixture
    def plugin(self):
        plugin = SqlAlchemyPluginAsync()
        plugin.init("dbname")
        return plugin

    @fixture
    def mget_my_settings(self, mocker, plugin):
        mock = mocker.patch.object(plugin, "get_my_settings")
        mock.return_value = {
            "url": "sqlite:///tmp.first.db",
            "options": {"optionkey": "option value"},
        }
        return mock

    def test_enter(self, plugin, masync_session):
        """
        .enter should create new database session and assign it to the context
        """
        mcontext = MagicMock()
        plugin.engine = MagicMock()

        assert plugin.enter(mcontext) == masync_session.return_value
        masync_session.assert_called_once_with(
            plugin.engine, expire_on_commit=False
        )

    def test_start(self, plugin, mapp, mcreate_async_engine, mget_my_settings):
        """
        .start should create proper sqlalchemy engine.
        """
        plugin.start(mapp)

        mcreate_async_engine.assert_called_once_with(
            "sqlite:///tmp.first.db", optionkey="option value"
        )



from contextvars import ContextVar
from unittest.mock import MagicMock
from unittest.mock import sentinel

from pytest import fixture

from qq.context import Context


class Fixtures:
    @fixture
    def mplugin(self):
        mock = MagicMock()
        mock.enter.return_value = sentinel.name
        return mock

    @fixture
    def mapp(self, mplugin):
        mock = MagicMock()
        mock.is_started = True
        mock.context = ContextVar("testing")
        mock.plugins = {
            "name": mplugin,
        }
        return mock

    @fixture
    def data(self):
        return {
            "name": 0,
        }

    @fixture
    def context(self, mapp):
        return Context(mapp)



from unittest.mock import MagicMock

from pytest import fixture


class PluginFixtures:
    @fixture
    def mconfigurator(self):
        return MagicMock()

    @fixture
    def mpyramid(self):
        return MagicMock()

    @fixture
    def mapplication(self):
        return MagicMock()



from copy import copy
from functools import update_wrapper
from inspect import iscoroutine
from inspect import iscoroutinefunction
from inspect import signature
from typing import Any
from typing import Callable

from qq.context import Context


class Inicjator:
    def init(self, context, arguments):
        self.context = context
        self.arguments = arguments

    def start(self) -> Any: ...  # pragma: no cover
    def finish(self, er=None): ...  # pragma: no cover


class ContextInicjator(Inicjator):
    def __init__(self, key: str):
        self.key = key

    def start(self):
        return self.context[self.key]


class FunctionWrapper:
    is_async = False

    def __init__(self, fun: Callable):
        self.application = None
        self.inicjators = {}
        self.fun = fun
        update_wrapper(self, fun)

    def __call__(self, *args, **kwargs):
        sig = signature(self.fun)
        arguments = sig.bind_partial(*args, **kwargs).arguments
        inicjators = []
        with Context(self.application) as context:
            for key in sig.parameters.keys():
                inicjator = self.inicjators.get(key)
                if key not in arguments and inicjator:
                    inicjator = copy(inicjator)
                    inicjators.append(inicjator)
                    inicjator.init(context, arguments)
                    arguments[key] = inicjator.start()

        try:
            error = None
            result = self.fun(**arguments)
        except Exception as er:
            error = er
            raise
        finally:
            for injector in reversed(inicjators):
                injector.finish(error)

        return result

    def set_application(self, application):
        self.application = application

    def set_injector(self, key, injector):
        self.inicjators[key] = injector

    def __repr__(self):
        return repr(self.fun)


class AsyncFunctionWrapper(FunctionWrapper):
    is_async = True

    async def __call__(self, *args, **kwargs):
        sig = signature(self.fun)
        arguments = sig.bind_partial(*args, **kwargs).arguments
        inicjators = []
        with Context(self.application) as context:
            for key in sig.parameters.keys():
                inicjator = self.inicjators.get(key)
                if key not in arguments and inicjator:
                    inicjator = copy(inicjator)
                    inicjators.append(inicjator)
                    inicjator.init(context, arguments)
                    iresult = inicjator.start()
                    if iscoroutine(iresult):
                        iresult = await iresult
                    arguments[key] = iresult

        try:
            error = None
            result = await self.fun(**arguments)
        except Exception as er:
            error = er
            raise
        finally:
            for injector in reversed(inicjators):
                iresult = injector.finish(error)
                if iscoroutine(iresult):
                    await iresult

        return result


def _init_function(fun):
    if isinstance(fun, FunctionWrapper):
        return fun

    if iscoroutinefunction(fun):
        return AsyncFunctionWrapper(fun)
    else:
        return FunctionWrapper(fun)


class SetApplication:
    def __init__(self, application):
        self.application = application

    def __call__(self, fun):
        fun = _init_function(fun)
        fun.set_application(self.application)
        return fun


class SetInicjator:
    def __init__(self, key, inicjator):
        self.key = key
        self.inicjator = inicjator

    def __call__(self, fun):
        fun = _init_function(fun)
        fun.set_injector(self.key, self.inicjator)
        return fun
